module CoolGirlChinaMMC1MMC3 # (
		parameter DOUBLE_WRITE_FIX = 1,				// block two writes in a row (RMW) for games like Snow Bros. and Bill & Ted's Excellent Adventure
		parameter USE_MAPPER_001 = 1,					// mapper #001 - MMC1
		parameter USE_MAPPER_002 = 0,					// mapper #002 - UxROM
		parameter USE_MAPPER_003 = 0,					// mapper #003 - CNROM
		parameter USE_MAPPER_004 = 1,					// mapper #004 - MMC3/MMC6
		parameter USE_MAPPER_007 = 1,					// mapper #007 - AxROM
		parameter USE_MAPPER_009_010 = 0,			// mapper #009 - MMC2, #10 - MMC4
		parameter USE_MAPPER_011 = 0,					// mapper #011 - Color Dreams
		parameter USE_MAPPER_066 = 0,					// mapper #066 - GxROM
		parameter USE_MAPPER_071 = 0,					// mapper #071 (for Fire Hawk only), require USE_MAPPER_002
		parameter USE_MAPPER_118 = 0,					// mapper #118 - TxSROM, require MAPPER_004
		parameter USE_MAPPER_189 = 0,					// mapper #189   require MAPPER_004
		parameter USE_MAPPER_241 = 0, 				// mapper #241 - BNROM, require USE_MAPPER_007

		parameter MMC3_BITSIZE = 8						// Maximum size for MMC3 PRG (6=512KB - standard size, 8=2MB - required for some hacks)
)
(
	input m2,
	input romsel,
	input cpu_rw_in,
	input [3:0] cpu_addr_in_low,
	input [14:8] cpu_addr_in,
	inout [7:0] cpu_data_in,
	output [26:13] cpu_addr_out,
	output flash_we,
	output flash_oe,
	output sram_ce,
	output sram_we,
	output sram_oe,

	input ppu_rd_in,
	input [13:0] ppu_addr_in,
	output [17:10] ppu_addr_out,
	output ppu_ciram_a10,

	output irq
);
	assign cpu_addr_out[26:13] = {cpu_base[26:14] | (cpu_addr_mapped[20:14] & ~prg_mask[20:14]), cpu_addr_mapped[13]};
	assign ppu_addr_out[17:10] = {ppu_addr_mapped[17:13] & ~chr_mask[17:13], ppu_addr_mapped[12:10]};

	assign cpu_data_in = cpu_data_out_enabled ? cpu_data_out : 8'bZZZZZZZZ;
	wire flash_ce_w = romsel;
	assign flash_oe = ~cpu_rw_in | flash_ce_w;

	assign flash_we = cpu_rw_in | flash_ce_w | ~prg_write_enabled;
	wire sram_ce_w = ~(cpu_addr_in[14] & cpu_addr_in[13] & m2 & romsel & sram_enabled);
	assign sram_ce = sram_ce_w;
	assign sram_we = cpu_rw_in | sram_ce_w;
	assign sram_oe = ~cpu_rw_in | sram_ce_w;


	reg [26:14] cpu_base = 0;
	reg [20:14] prg_mask = 7'b1111000;
	reg [17:13] chr_mask = 0;
	reg [2:0] prg_mode = 0;
	reg [7:0] prg_bank_a = 0;
	reg [7:0] prg_bank_b = 1;
	reg [7:0] prg_bank_c = 8'b11111110;
	reg [7:0] prg_bank_d = 8'b11111111;
	reg [2:0] chr_mode = 0;
	reg [7:0] chr_bank_a = 0;
	reg [7:0] chr_bank_b = 1;
	reg [7:0] chr_bank_c = 2;
	reg [7:0] chr_bank_d = 3;
	reg [7:0] chr_bank_e = 4;
	reg [7:0] chr_bank_f = 5;
	reg [7:0] chr_bank_g = 6;
	reg [7:0] chr_bank_h = 7;
	reg [4:0] mapper = 0;
	reg [2:0] flags = 0;
	reg sram_enabled = 0;
	reg prg_write_enabled = 0;
	reg [1:0] mirroring = 0;

	// some common registers for all mappers
	reg [7:0] r1 = 0;
	reg [4:0] mapper1_shift_register = 0;

	// for scanline-based interrupts
	reg [7:0] irq_scanline_counter = 0;
	reg [2:0] a12_low_time = 0;
	reg irq_scanline_reload = 0;
	reg [7:0] irq_scanline_latch = 0;
	reg irq_scanline_reload_clear = 0;
	reg irq_scanline_enabled = 0;
	reg irq_scanline_ready = 0;
	reg irq_scanline_out = 0;

	// for MMC2/MMC4
	reg ppu_latch0 = 0;
	reg ppu_latch1 = 0;

	reg readed;

	wire cpu_data_out_enabled;
	wire [7:0] cpu_data_out;
	assign cpu_data_out_enabled =
		(m2 & romsel & cpu_rw_in & ~(mapper[4] | mapper[3] | mapper[2] | mapper[2] | mapper[0]) & (cpu_addr_c000 & cpu_addr_in[12])) ? 1'b1 : 1'b0;
	assign cpu_data_out = 8'b00000000;
		

	assign ppu_ciram_a10 = (USE_MAPPER_118 & (mapper == 5'b10100) & flags[0]) ? ppu_addr_mapped[17] :
		(mirroring[1] ? mirroring[0] : (mirroring[0] ? ppu_addr_in[11] : ppu_addr_in[10])); // vertical / horizontal, 1Sa, 1Sb			


	wire [20:13] cpu_addr_mapped = // (map_rom_on_6000 & romsel & m2) ? prg_bank_6000 :
	(
		prg_mode[2] ? (
			prg_mode[1] ? (				
				prg_mode[0] ? (
					// 111 - 0x8000(A)
					{prg_bank_a[7:2], cpu_addr_in[14:13]}
				) : (
					// 110 - 0x8000(B)
					{prg_bank_b[7:2], cpu_addr_in[14:13]}
				)
			) : ( // prg_mode[1]
				prg_mode[0] ? (
					// 101 - 0x2000(C)+0x2000(B)+0x2000(A)+0x2000(D)
					cpu_addr_in[14] ? (cpu_addr_in[13] ? prg_bank_d : prg_bank_a) : (cpu_addr_in[13] ? prg_bank_b : prg_bank_c)
				) : ( // prg_mode[0]
					// 100 - 0x2000(A)+0x2000(B)+0x2000(C)+0x2000(D)
					cpu_addr_in[14] ? (cpu_addr_in[13] ? prg_bank_d : prg_bank_c) : (cpu_addr_in[13] ? prg_bank_b : prg_bank_a)
				)
			)
		) : ( // prg_mode[2]
			prg_mode[0] ? (
				// 0x1 - 0x4000(C) + 0x4000 (A)
				{cpu_addr_in[14] ? prg_bank_a[7:1] : prg_bank_c[7:1], cpu_addr_in[13]}
			) : ( // prg_mode[0]
				// 0x0 - 0x4000(A) + 0x4000 (С)
				{cpu_addr_in[14] ? prg_bank_c[7:1] : prg_bank_a[7:1], cpu_addr_in[13]}
			)
		)
	);

	wire [17:10] ppu_addr_mapped = chr_mode[2] ? (
		chr_mode[1] ? (
			chr_mode[0] ? ( 
				// 111 - 0x400(A)+0x400(B)+0x400(C)+0x400(D)+0x400(E)+0x400(F)+0x400(G)+0x400(H)
				ppu_addr_in[12] ? 
					(ppu_addr_in[11] ? (ppu_addr_in[10] ? chr_bank_h : chr_bank_g) : 
					(ppu_addr_in[10] ? chr_bank_f : chr_bank_e)) : (ppu_addr_in[11] ? (ppu_addr_in[10] ? chr_bank_d : chr_bank_c) : (ppu_addr_in[10] ? chr_bank_b : chr_bank_a))
			) : ( // chr_mode[0]
				// 110 - 0x800(A)+0x800(C)+0x800(E)+0x800(G)
				{ppu_addr_in[12] ? 
					(ppu_addr_in[11] ? chr_bank_g[7:1] : chr_bank_e[7:1]) : 
					(ppu_addr_in[11] ? chr_bank_c[7:1] : chr_bank_a[7:1]), ppu_addr_in[10]}
			)
		) : ( // chr_mode[1]
			// 100 - 0x1000(A) + 0x1000(E)
			// 101 - 0x1000(A/B) + 0x1000(E/F) - MMC2 и MMC4
			{ppu_addr_in[12] ?
				(((USE_MAPPER_009_010) && chr_mode[0] && ppu_latch1) ? chr_bank_f[7:2] : chr_bank_e[7:2]) : 
				(((USE_MAPPER_009_010) && chr_mode[0] && ppu_latch0) ? chr_bank_b[7:2] : chr_bank_a[7:2]),
			ppu_addr_in[11:10]}
		)
	) : ( // chr_mode[2]
		chr_mode[1] ? (
			// 010 - 0x800(A)+0x800(C)+0x400(E)+0x400(F)+0x400(G)+0x400(H) 
			// 011 - 0x400(E)+0x400(F)+0x400(G)+0x400(H)+0x800(A)+0x800(С)
			(ppu_addr_in[12]^chr_mode[0]) ? 
				(ppu_addr_in[11] ?
					(ppu_addr_in[10] ? chr_bank_h : chr_bank_g) : 
					(ppu_addr_in[10] ? chr_bank_f : chr_bank_e)
				) : (
					ppu_addr_in[11] ? {chr_bank_c[7:1],ppu_addr_in[10]} : {chr_bank_a[7:1],ppu_addr_in[10]}
				)
		) : ( // chr_mode[1]			
				// 000 - 0x2000(A)
				{chr_bank_a[7:3], ppu_addr_in[12:10]}
			 )
	);

	wire cpu_addr_8000 = ~(cpu_addr_in[14] | cpu_addr_in[13]);	// 00
	wire cpu_addr_a000 = ~cpu_addr_in[14] & cpu_addr_in[13];		// 01
	wire cpu_addr_c000 = cpu_addr_in[14] & ~cpu_addr_in[13];		// 10
	wire cpu_addr_e000 = cpu_addr_in[14] & cpu_addr_in[13];		// 11

	assign irq = (irq_scanline_out) ? 1'b0 : 1'bZ;

	always @ (negedge m2)
	begin
		if (DOUBLE_WRITE_FIX)
			readed <= cpu_rw_in;

		if (!cpu_rw_in) // write
		begin
			if (romsel) // $0000-$7FFF
			begin
				if (cpu_addr_c000 & cpu_addr_in[12]) // $5000-5FFF & lockout is off
				begin
					case (cpu_addr_in_low[2:0]) // case (cpu_addr_in[2:0])
						3'b000: // $5xx0
							{cpu_base[26:22]} <= cpu_data_in[4:0];						// CPU base address A26-A22
						3'b001: // $5xx1
							cpu_base[21:14] <= cpu_data_in[7:0]; 						// CPU base address A21-A14
						3'b010: // $5xx2
							prg_mask[20:14] <= cpu_data_in[6:0]; 						// CPU mask A18-A14
						3'b011: // $5xx3
							{prg_mode[2:0], chr_bank_a[7:3]} <= cpu_data_in[7:0];	// PRG mode, direct chr_bank_a access
						3'b100: // $5xx4
							{chr_mode[2:0], chr_mask[17:13]} <= cpu_data_in[7:0];	// CHR mode, CHR mask A17-A13
						3'b101: // $5xx5
							//{prg_bank_a[5:1], sram_page[1:0]} = cpu_data_in[6:0]; // direct prg_bank_a access, current SRAM page 0-3
							{prg_bank_a[5:1]} <= cpu_data_in[6:2]; // direct prg_bank_a access
						3'b110: // $5xx6
							{flags[2:0], mapper[4:0]} <= cpu_data_in[7:0];			// some flags, mapper
						3'b111: // $5xx7														
							// some other parameters						
							//{lockout, four_screen, mirroring[1:0], prg_write_enabled, chr_write_enabled, sram_enabled} = {cpu_data_in[7], cpu_data_in[5:0]};
							{mirroring[1:0], prg_write_enabled, sram_enabled} <= {cpu_data_in[4:2], cpu_data_in[0]};
					endcase

					if (USE_MAPPER_009_010 && mapper == 5'b10001) prg_bank_b <= 8'b11111101;
				end

				// Mapper #189
				// It's MMC3 with flag1
				if (USE_MAPPER_189 & flags[1] & (mapper == 5'b10100))
				begin
					//if (cpu_addr_in[14:0] >= 15'h4120) // $4120-$7FFF
					if (cpu_addr_in[14:8] >= 7'h41) // $4120-$7FFF
					begin
						prg_bank_a[5:2] <= cpu_data_in[3:0] | cpu_data_in[7:4];
					end
				end				

			end else begin // $8000-$FFFF

				// Mapper #2 - UxROM
				// flag0 - mapper #71 - for Fire Hawk only.
				// other mapper-#71 games are UxROM
				if (USE_MAPPER_002 && mapper == 5'b00001)
				begin
					if (!USE_MAPPER_071 | ~flags[0] | (cpu_addr_in[14:12] != 3'b001))
					begin
						prg_bank_a[5:1] <= cpu_data_in[4:0];
					end else begin // CodeMasters, blah. Mirroring control used only by Fire Hawk
						mirroring[1:0] <= {1'b1, cpu_data_in[4]};
					end
				end
				
				// Mapper #3 - CNROM
				if (USE_MAPPER_003 && mapper == 5'b00010)
				begin
					chr_bank_a[7:3] <= cpu_data_in[4:0];
				end

				// Mapper #7 - AxROM, mapper #241 - BNROM
				if (USE_MAPPER_007 && mapper == 5'b01000)
				begin
					prg_bank_a[5:2] <= cpu_data_in[3:0];
					if (!USE_MAPPER_241 || !flags[0]) // BNROM?
						mirroring <= {1'b1, cpu_data_in[4]};
				end

				// Mapper #11 - ColorDreams
				if (USE_MAPPER_011 && mapper == 5'b01010)
				begin
					prg_bank_a[3:2] <= cpu_data_in[1:0];
					chr_bank_a[6:3] <= cpu_data_in[7:4];
				end

				// Mapper #66 - GxROM
				if (USE_MAPPER_066 && mapper == 5'b01011)
				begin
					prg_bank_a[3:2] <= cpu_data_in[5:4];
					chr_bank_a[4:3] <= cpu_data_in[1:0];					
				end
				
				// Mapper #1 - MMC1
				if (USE_MAPPER_001 && mapper == 5'b10000 && (!DOUBLE_WRITE_FIX || readed))
				begin
					mapper1_shift_register <= (cpu_data_in[7] | mapper1_shift_register[0]) ? 5'b10000
						: {cpu_data_in[0], mapper1_shift_register[4:1]};

					if (mapper1_shift_register[0]) // last bit in cpu_data_in[0] for sync mode
					begin
						if (cpu_addr_8000) // $8000-$9FFF
						begin
							prg_mode <= mapper1_shift_register[4] ? {2'b00, ~mapper1_shift_register[3]} : 3'b111;
							prg_bank_c[4:0] <= mapper1_shift_register[3] ? 5'b11110 : 5'b00000;
							chr_mode <= {cpu_data_in[0], 2'b00};
							mirroring <= {~mapper1_shift_register[2], mapper1_shift_register[1]};
						end

						if (cpu_addr_a000) // $A000-$BFFF
						begin
							chr_bank_a[6:2] <= {cpu_data_in[0], mapper1_shift_register[4:1]};
							prg_bank_a[5] <= cpu_data_in[0]; // for SUROM, 512k PRG support
							prg_bank_c[5] <= cpu_data_in[0]; // for SUROM, 512k PRG support
						end

						if (cpu_addr_c000) // $C000-$DFFF
							chr_bank_e[6:2] <= {cpu_data_in[0], mapper1_shift_register[4:1]};

						if (cpu_addr_e000) // $E000-$FFFF
						begin
							prg_bank_a[4:1] <= mapper1_shift_register[4:1]; // $E000-$FFFF
							sram_enabled <= ~cpu_data_in[0];
						end
					end

					if (cpu_data_in[7]) // reset
					begin
						prg_mode <= 3'b000;	// 0x4000 (A) + fixed last (C)
						prg_bank_c[4:0] <= 5'b11110;
					end
				end
				
				// Mapper #9 and #10 - MMC2 and MMC4
				// flag0 - 0=MMC2, 1=MMC4
				if (USE_MAPPER_009_010 && mapper == 5'b10001)
				begin
					case (cpu_addr_in[14:12])
						3'b010: if (~flags[0]) // $A000-$AFFF
										prg_bank_a[3:0] <= cpu_data_in[3:0];
									else
										prg_bank_a[4:1] <= cpu_data_in[3:0];
						3'b011: chr_bank_a[6:2] <= cpu_data_in[4:0]; // $B000-$BFFF
						3'b100: chr_bank_b[6:2] <= cpu_data_in[4:0]; // $C000-$CFFF
						3'b101: chr_bank_e[6:2] <= cpu_data_in[4:0]; // $D000-$DFFF
						3'b110: chr_bank_f[6:2] <= cpu_data_in[4:0]; // $E000-$EFFF
						3'b111: mirroring <= {1'b0, cpu_data_in[0]}; // $F000-$FFFF
					endcase
				end
	
				// Mapper #4 - MMC3/MMC6
				/*
				r0[2:0] - internal register
				flag0 - TxSROM
				flag1 - mapper #189
				*/				
				if (USE_MAPPER_004 && mapper == 5'b10100)
				begin
					if (cpu_addr_8000 & ~cpu_addr_in_low[0]) // $8000-$9FFE, even
					begin
						r1[2:0] <= cpu_data_in[2:0];
						if (!USE_MAPPER_189 | ~flags[1])
							prg_mode <= {2'b10, cpu_data_in[6]};
						chr_mode <= {2'b01, cpu_data_in[7]};
					end

					if (cpu_addr_8000 & cpu_addr_in_low[0]) // $8001-$9FFF, odd
					begin
						case (r1[2:0])
							3'b000: chr_bank_a <= cpu_data_in;
							3'b001: chr_bank_c <= cpu_data_in;
							3'b010: chr_bank_e <= cpu_data_in;
							3'b011: chr_bank_f <= cpu_data_in;
							3'b100: chr_bank_g <= cpu_data_in;
							3'b101: chr_bank_h <= cpu_data_in;
							3'b110: if (!USE_MAPPER_189 | ~flags[1]) prg_bank_a[(MMC3_BITSIZE-1):0] <= cpu_data_in[(MMC3_BITSIZE-1):0];
							3'b111: if (!USE_MAPPER_189 | ~flags[1]) prg_bank_b[(MMC3_BITSIZE-1):0] <= cpu_data_in[(MMC3_BITSIZE-1):0];
						endcase
					end

					if (cpu_addr_a000 & ~cpu_addr_in_low[0])
						mirroring <= {1'b0, cpu_data_in[0]}; // $A000-$BFFE, even (mirroring)

					if (cpu_addr_c000 & ~cpu_addr_in_low[0])
						irq_scanline_latch <= cpu_data_in; // $C000-$DFFE, even (IRQ latch)

					if (cpu_addr_c000 & cpu_addr_in_low[0])
						irq_scanline_reload <= 1; // $C001-$DFFF, odd

					if (cpu_addr_e000)
						irq_scanline_enabled <= cpu_addr_in_low[0]; // $E000-$FFFF
				end
				
			end // romsel
		end // write

		// some IRQ stuff
		if (irq_scanline_reload_clear)
			irq_scanline_reload <= 0;

		irq_scanline_ready <= irq_scanline_enabled & ~irq_scanline_counter_zero;

		// Fire scanline IRQ if counter is zero
		// BUT doesn't fire it when it's zero end reenabled
		if (~irq_scanline_enabled)
			irq_scanline_out <= 0;
		else if (irq_scanline_ready & irq_scanline_counter_zero)
			irq_scanline_out <= 1;
	end

	wire irq_scanline_counter_zero = ~(irq_scanline_counter[7] | irq_scanline_counter[6] | irq_scanline_counter[5] | irq_scanline_counter[4] | irq_scanline_counter[3] | irq_scanline_counter[2] | irq_scanline_counter[1] | irq_scanline_counter[0]);

	// IRQ counter
	always @ (posedge ppu_addr_in[12])
	begin
		irq_scanline_reload_clear <= a12_low_time[2] ? irq_scanline_reload : irq_scanline_reload_clear; // prime reload flag or clear it
		irq_scanline_counter <= (a12_low_time[2] & (irq_scanline_reload | irq_scanline_counter_zero)) ? irq_scanline_latch : irq_scanline_counter - a12_low_time[2];
	end

	// A12 must be low for 3 rises of M2
	always @ (posedge m2, posedge ppu_addr_in[12])
	begin
		a12_low_time <= ppu_addr_in[12] ? 3'b000 : {a12_low_time[1:0], 1'b1}; // shift register
	end

	// for MMC2/MMC4
	always @ (negedge ppu_rd_in)
	begin
		if (USE_MAPPER_009_010)
		begin
			if (ppu_addr_in[13:3] == 11'b00111111011) ppu_latch0 <= 0;
			if (ppu_addr_in[13:3] == 11'b00111111101) ppu_latch0 <= 1;
			if (ppu_addr_in[13:3] == 11'b01111111011) ppu_latch1 <= 0;
			if (ppu_addr_in[13:3] == 11'b01111111101) ppu_latch1 <= 1;
		end
	end

endmodule
